@@ -355,11 +355,17 @@ static int pcm_digital_in_open(snd_pcm_substream_t *substream)
 	if (err < 0)
 		goto din_exit;
 
-	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_CHANNELS,
-					hw_rule_capture_channels_by_format, 0, SNDRV_PCM_HW_PARAM_FORMAT, -1)) < 0)
+	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0,
+				       SNDRV_PCM_HW_PARAM_CHANNELS,
+				       hw_rule_capture_channels_by_format,
+				       NULL, SNDRV_PCM_HW_PARAM_FORMAT,
+				       -1)) < 0)
 		goto din_exit;
-	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_FORMAT,
-					hw_rule_capture_format_by_channels, 0, SNDRV_PCM_HW_PARAM_CHANNELS, -1)) < 0)
+	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0,
+				       SNDRV_PCM_HW_PARAM_FORMAT,
+				       hw_rule_capture_format_by_channels,
+				       NULL, SNDRV_PCM_HW_PARAM_CHANNELS,
+				       -1)) < 0)
 		goto din_exit;
 
 	atomic_inc(&chip->opencount);
@@ -391,11 +397,17 @@ static int pcm_digital_out_open(snd_pcm_substream_t *substream)
 	if (err < 0)
 		goto dout_exit;
 
-	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_CHANNELS,
-					hw_rule_playback_channels_by_format, 0, SNDRV_PCM_HW_PARAM_FORMAT, -1)) < 0)
+	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0,
+				       SNDRV_PCM_HW_PARAM_CHANNELS,
+				       hw_rule_playback_channels_by_format,
+				       NULL, SNDRV_PCM_HW_PARAM_FORMAT,
+				       -1)) < 0)
 		goto dout_exit;
-	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_FORMAT,
-					hw_rule_playback_format_by_channels, 0, SNDRV_PCM_HW_PARAM_CHANNELS, -1)) < 0)
+	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0,
+				       SNDRV_PCM_HW_PARAM_FORMAT,
+				       hw_rule_playback_format_by_channels,
+				       NULL, SNDRV_PCM_HW_PARAM_CHANNELS,
+				       -1)) < 0)
 		goto dout_exit;
 	atomic_inc(&chip->opencount);
 	if (atomic_read(&chip->opencount) > 1 && chip->rate_set)
@@ -1785,7 +1797,7 @@ static __devinit int snd_echo_create(snd_card_t *card, struct pci_dev *pci, echo
 	pci_set_master(pci);
 
 	/* allocate a chip-specific data */
-	chip = kcalloc(1, sizeof(echoaudio_t), GFP_KERNEL);
+	chip = kzalloc(sizeof(*chip), GFP_KERNEL);
 	if (!chip) {
 		pci_disable_device(pci);
 		return -ENOMEM;
@@ -1878,7 +1890,7 @@ static int __devinit snd_echo_probe(struct pci_dev *pci, const struct pci_device
 		return err;
 	}
 
-	strcpy(card->driver, "Echoaudio " ECHOCARD_NAME);
+	strcpy(card->driver, "Echo_" ECHOCARD_NAME);
 	strcpy(card->shortname, chip->card_name);
 
 	dsp = "56301";
